/* File t2f.c
 * October 10, 2002
 * By Doug Mink, Harvard-Smithsonian Center for Astrophysics
 * Send bug reports to dmink@cfa.harvard.edu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <math.h>
#include "libtiff/tiffio.h"
#include "libwcs/fitsfile.h"

#define TIFF_RED	1	/* Read only red sample */
#define TIFF_GREEN	2	/* Read only green sample */
#define TIFF_BLUE	3	/* Read only blue sample */
#define TIFF_RG		4	/* Write mean of red+green samples */
#define TIFF_RB		5	/* Write mean of red+blue samples */
#define TIFF_GB		6	/* Write mean of blue+green samples */
#define TIFF_RGB	7	/* Write mean of three colors */
#define MAXRANGE 20

/* Structure for dealing with ranges */
struct Range {
    double first;	/* Current minimum value */
    double last;	/* Current maximum value */
    double step;	/* Current step in value */
    double value;	/* Current value */
    double ranges[MAXRANGE*3];  /* nranges sets of first, last, step */
    int nvalues;	/* Total number of values in all ranges */
    int nranges;	/* Number of ranges */
    int irange;		/* Index of current range */
};

/* Subroutines for dealing with ranges */
static struct Range *RangeInit(); /* Initialize range structure from string */
static int isrange();	/* Return 1 if string is a range of numbers, else 0 */
static int rgetn();	/* Return number of values in all ranges */
static int rgeti4();	/* Return next number in range as integer */
static double rgetr8();	/* Return next number in range as double */
static void rstart();	/* Restart range */

static void usage();
static void TIFFtoFITS ();

static int verbose = 0;		/* verbose/debugging flag */
static char *RevMsg = "T2F 3.2.0, 10 October 2002, Doug Mink (dmink@cfa.harvard.edu)";
static int version = 0;		/* If 1, print only program name and version */
static char outname[128];	/* Name for output image */
static char outdir[256];	/* Output directory pathname */
static int first = 1;
static int nlog = 100;
static int negative = 1;
static int color = TIFF_RGB;
static int limitval = 0;
static int background = 0;
static int bitshift = 0;
static int xcpix = 0;
static int ycpix = 0;
static int xdpix = 0;
static int ydpix = 0;
static char *rrange;	/* Row range string */
static char *crange;	/* Column range string */

main (ac, av)
int ac;
char **av;
{
    char *str;
    char *listfile;
    char *fn;
    char filename[256];
    char temp[80];
    int ifile, nfile;
    FILE *flist;
    char *nextarg;

    crange = NULL;
    rrange = NULL;
    fn = NULL;
    outname[0] = 0;
    outdir[0] = 0;

    /* Check for help or version command first */
    str = *(av+1);
    if (!str || !strcmp (str, "help") || !strcmp (str, "-help"))
	usage("");
    if (!strcmp (str, "version") || !strcmp (str, "-version")) {
	version = 1;
	usage("");
	}

    /* crack arguments */
    for (av++; --ac > 0; av++) {

	/* Read command */
	if (*(str = *av) == '-') {
	    char c;
	    while (c = *++str) {
		switch (c) {
		case '1':	/* shift lowest bit */
		    bitshift++;
		    break;
		case 'v':	/* more verbosity */
		    verbose++;
		    break;
		case 'b':	/* read only blue samples */
		    if (color == TIFF_RED)
			color = TIFF_RB;
		    else if (color == TIFF_GREEN)
			color = TIFF_GB;
		    else if (color == TIFF_RG)
			color = TIFF_RGB;
		    else
			color = TIFF_BLUE;
		    break;
		case 'c':	/* center coordinates for extraction */
		    if (ac < 3)
			usage ("-c needs two arguments");
		    xcpix = atoi (*++av);
		    ac--;
		    ycpix = atoi (*++av);
		    ac--;
		    break;
		case 'd':	/* output directory */
		    if (ac < 2)
			usage ("-d needs an argument");
		    strcpy (outdir, *++av);
		    ac--;
		    break;
		case 'g':	/* read green samples */
		    if (color == TIFF_RED)
			color = TIFF_RG;
		    else if (color == TIFF_BLUE)
			color = TIFF_GB;
		    else if (color == TIFF_RB)
			color = TIFF_RGB;
		    else
			color = TIFF_GREEN;
		    break;
		case 'i':	/* logging interval */
		    if (ac < 2)
			usage ("-i needs an argument");
		    nlog = atoi (*++av);
		    ac--;
		    break;
		case 'k':	/* background pixel value */
		    if (ac < 2)
			usage ("-k needs an argument");
		    background = atoi (*++av);
		    ac--;
		    break;
		case 'l':	/* limiting pixel value */
		    if (ac < 2)
			usage ("-l needs an argument");
		    limitval = atoi (*++av);
		    ac--;
		    break;
		case 'm':	/* use mean of multiple samples */
		    color = TIFF_RGB;
		    break;
		case 'n':	/* image is a negative */
		    negative = 1;
		    break;
		case 'o':	/* output file name */
		    if (ac < 2)
			usage ("-o needs an argument");
		    strcpy (outname, *++av);
		    ac--;
		    break;
		case 'p':	/* image is a positive */
		    negative = 0;
		    break;
		case 'r':	/* read only red samples */
		    if (color == TIFF_GREEN)
			color = TIFF_RG;
		    else if (color == TIFF_BLUE)
			color = TIFF_RB;
		    else if (color == TIFF_GB)
			color = TIFF_RGB;
		    else
			color = TIFF_RED;
		    break;
		case 's':	/* output to stdout */
		    strcpy (outname, "stdout");
		    break;
		case 'x':	/* width and height for extraction */
		    if (ac < 2)
			usage ("-x needs at least one argument");
		    xdpix = atoi (*++av);
		    ac--;
		    if (ac > 1) {
			nextarg = *(av+1);
			if (isnum (nextarg)) {
			    ydpix = atoi (nextarg);
			    av++;
			    ac--;
			    }
			else
			    ydpix = xdpix;
			}
		    else
			ydpix = xdpix;
		    break;
	        default:
		    sprintf (temp, "Illegal argument '%c'", c);
		    usage(temp);
		    break;
		}
		}
    	    }

        /* center and size of section to extract */
        else if (isnum (str)) {
	    if (!xcpix)
		xcpix = atoi (str);
	    else if (!ycpix)
		ycpix = atoi (str);
	    else if (!xdpix)
		xdpix = atoi (str);
	    else if (!ydpix)
		ydpix = atoi (str);
	    }

        /* range of pixels to extract */
        else if (isrange (str)) {
	    if (crange == NULL)
		crange = str;
	    else
		rrange = str;
	    }

 	else if (*av[0] == '@') {
	    listfile = *av + 1;
	    }

        /* file name */
        else
            fn = str;
	}

    /* If center is set, but not size, extract 500x500 image */
    if (!crange && xcpix && ycpix && xdpix == 0) {
	xdpix = 500;
	ydpix = 500;
	}

    /* If one side is set, set the other */
    if (xdpix && !ydpix)
	ydpix = xdpix;

    /* now there are ac remaining file names starting at av[0] */
    if (listfile && isimlist (listfile)) {
	nfile = getfilelines (listfile);
	if ((flist = fopen (listfile, "r")) == NULL) {
	    fprintf (stderr,"I2F: Image list file %s cannot be read\n",
		     listfile);
	    usage ();
	    }
	for (ifile = 0; ifile < nfile; ifile++) {
	    first_token (flist, 254, filename);
	    TIFFtoFITS (filename);
	    }
	fclose (flist);
	}

    else if (fn) {
	TIFFtoFITS (fn);
	}

    return (0);
}

static void
usage (errmsg)

char *errmsg;	/* Error message */
{
    fprintf (stderr,"%s\n",RevMsg);
    if (version)
	exit (-1);
    if (*errmsg)
	fprintf (stderr, "*** %s ***\n", errmsg);
    fprintf (stderr,"Write FITS files from TIFF image files\n");
    fprintf(stderr,"Usage: t2f -sv[r|g|b|m] [-i num] [-o name] [-d dir] file.tif ...\n");
    fprintf(stderr,"  or : t2f [-sv1][-i num][-c num num][-x num[num]] [xrange yrange] [-o name] [-d path] @tifflist\n");
    fprintf(stderr,"  xrange: Columns to extract in format x1-x2\n");
    fprintf(stderr,"  yrange: Rows to extract in format y1-y2\n");
    fprintf(stderr,"  -1: Shift out lowest order bit for signed 16-bit output\n");
    fprintf(stderr,"  -b: use blue from TIFF file\n");
    fprintf(stderr,"  -c x y: Extract around this center TIFF coordinate\n");
    fprintf(stderr,"  -d dir: write FITS file(s) to this directory\n");
    fprintf(stderr,"  -g: use green from TIFF file\n");
    fprintf(stderr,"  -i num: log rows as they are copied at this interval\n");
    fprintf(stderr,"  -m: Write mean of TIFF file red, green, and blue (default)\n");
    fprintf(stderr,"  -k num: Add this constant to every pixel\n");
    fprintf(stderr,"  -l num: Set all values greater than this to this number\n");
    fprintf(stderr,"  -n: Convert negative TIFF file to positive (default)\n");
    fprintf(stderr,"  -o name: output name for one file\n");
    fprintf(stderr,"  -p: TIFF file is already positive\n");
    fprintf(stderr,"  -r: use red from TIFF file\n");
    fprintf(stderr,"  -s: write output to standard output\n");
    fprintf(stderr,"  -x dx[,dy]; eXtract dx * dx or dx * dy pixels (500 default)\n");
    fprintf(stderr,"  -v: verbose\n");
    exit (1);
}

static void
TIFFtoFITS (name)
char *name;
{
    char *header;	/* FITS header */
    int lhead;		/* Maximum number of bytes in FITS header */
    int nbhead;		/* Actual number of bytes in FITS header */
    char pixname[256];	/* Pixel file name */
    char history[128];	/* for HISTORY line */
    char *filename;	/* Pointer to start of file name */
    char tiffname[256];	/* Name of TIFF file */
    char fitsname[256];	/* Name of FITS file */
    char fitspath[256];	/* Pathname of FITS file  */
    char *hdrfile1;
    char *fitsfile;
    char *image;        /* FITS image */
    int nbytes;
    double cra;         /* Center right ascension in degrees (returned) */
    double cdec;        /* Center declination in degrees (returned) */
    double dra;         /* Right ascension half-width in degrees (returned) */
    double ddec;        /* Declination half-width in degrees (returned) */
    double secpix;      /* Arcseconds per pixel (returned) */
    int wp;             /* Image width in pixels (returned) */
    int hp;             /* Image height in pixels (returned) */
    int sysout=0;       /* Coordinate system to return (0=image, returned) */
    double eqout=0.0;   /* Equinox to return (0=image, returned) */
    int i, j;
    int irow1, irow2;	/* First and last rows to read */
    int icol1, icol2;	/* First and last columns to read */
    int ifrow1, ifrow2;	/* First and last rows to write */
    int ifcol1, ifcol2;	/* First and last columns to write */
    int ncolw;		/* Number of columns to write */
    int nbimage;        /* Number of bytes in image */
    int nbskip;         /* Number of bytes to skip in file before image */
    int nbfile;         /* Number of bytes in file */
    int nbread;         /* Number of bytes actually read from file */
    char command[256];
    char *ext;		/* Pointer to start of extension */
    char *endchar;
    char *ltime;
    int nc, ier, bitpix;
    int fdw;
    char *bufout1;
    ushort *bufin2, *bufout2, *bufout2u;
    short *bufout2s;
    char *buf1;
    int tpix, bytepix, nprow;
    int nbrow, nblock, nbleft, sampix, nbout;
    uint16 nbsamp, nspix;
    uint32 nrows, ncols;
    tdata_t buf;
    short *bufout;
    short *bufoutf2;
    char *bufoutf1;
    char *bufin1;
    char temp[80];
    uint32 row;
    int irow;
    int vpix;
    int xdim = 1;
    int ydim = 1;
    int npfrst, nplast;
    int itemp;
    TIFF *tif;
    char ib0 = (char) 0;
    struct Range *xrange;    /* Column range structure */
    struct Range *yrange;    /* Row range structure */

    /* Open TIFF image */
    tif = TIFFOpen (name, "r");

    /* If found, make up output FITS file name */
    if (tif) {
	filename = strrchr (name,'/');
	if (filename)
	    filename = filename + 1;
	else
	    filename = name;
	strcpy (fitsname, filename);
	ext = strsrch (fitsname,".tif");
	if (ext)
	    *ext = (char) 0;
	if (verbose && first) {
	    fprintf (stderr,"%s\n",RevMsg);
	    fprintf (stderr, "Read TIFF image file %s", name);
	    }
	}

    /* Add .tif extension to make TIFF header file name if not present */
    else {
	strcpy (tiffname, name);
	strcat (tiffname,".tif");
	tif = TIFFOpen (tiffname, "r");
	if (!tif) {
	    fprintf (stderr,"Cannot read TIFF image file %s\n", tiffname);
	    return;
	    }
	filename = strrchr (tiffname,'/');
	if (filename)
	    filename = filename + 1;
	else
	    filename = tiffname;
	strcpy (fitsname, filename);
	if (verbose && first) {
	    fprintf (stderr,"%s\n",RevMsg);
	    fprintf (stderr,"Read TIFF image file %s\n", tiffname);
	    }
	}

    /* Initialize FITS header */
    lhead = 2880;
    header = (char *) calloc (1, lhead);
    for (i = 0; i < 2880; i++)
	header[i] = ' ';
    strcpy (header, "END ");
    hlength (header, 2880);
    hputl (header, "SIMPLE", 1);

    /* Set pixel size (negate for unsigned) */
    TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &nbsamp);
    bitpix = nbsamp;
    bytepix = nbsamp / 8;
    hputi4 (header, "BITPIX", bitpix);
    if (bytepix == 2)
	bitshift = 1;

    /* Set samples per pixel */
    TIFFGetField (tif, TIFFTAG_SAMPLESPERPIXEL, &nspix);
    sampix = nspix;
    if (verbose) {
	if (sampix > 1)
	    fprintf (stderr, " (%d bits, %d samples per pixel)", bitpix, sampix);
	else
	    fprintf (stderr, " (%d bits, %d sample per pixel)", bitpix, sampix);
	}

    if (color == TIFF_RED)
	strcat (fitsname, "r");
    else if (color == TIFF_GREEN)
	strcat (fitsname, "g");
    else if (color == TIFF_BLUE)
	strcat (fitsname, "b");
    else if (sampix > 1)
	strcat (fitsname, "m");
    strcat (fitsname, ".fits");

    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &ncols);
    TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &nrows);


    /* Set up limiting rows to read */
    if (rrange) {
	yrange = RangeInit (rrange, ydim);
	ifrow1 = yrange->ranges[0];
	ifrow2 = yrange->ranges[1];
	}
    else if (xcpix && ycpix) {
	int ydpix2 = ydpix / 2;
	ifrow1 = ycpix - ydpix2 + 1;
	ifrow2 = ycpix + ydpix2;
	}
    else{
	ifrow1 = 1;
	ifrow2 = nrows;
	}
    irow1 = nrows - ifrow2;
    irow2 = nrows - ifrow1;
    if (irow1 > nrows - 1)
	irow1 = nrows - 1;
    else if (irow1 < 0)
	irow1 = 0;
    if (irow2 > nrows - 1)
	irow2 = nrows - 1;
    else if (irow2 < 0)
	irow2 = 0;
    hp = irow2 - irow1 + 1;

    /* Set up first and number of columns to write */
    if (crange) {
	xrange = RangeInit (crange, xdim);
	ifcol1 = xrange->ranges[0];
	ifcol2 = xrange->ranges[1];
	}
    else if (xdpix) {
	int xdpix2 = xdpix / 2;
	ifcol1 = xcpix - xdpix2 + 1;
	ifcol2 = xcpix + xdpix2;
	}
    else {
	ifcol1 = 1;
	ifcol2 = ncols;
	}
    icol1 = ncols - ifcol2 + 1;
    icol2 = ncols - ifcol1 + 1;
    if (icol1 > ncols)
	icol1 = ncols;
    if (icol1 < 1)
	icol1 = 1;
    if (icol2 > ncols)
	icol2 = ncols;
    if (icol2 < 1)
	icol2 = 1;
    wp = icol2 - icol1 + 1;

    hputi4 (header, "NAXIS", 2);
    hputi4 (header, "NAXIS1", wp);
    hputi4 (header, "NAXIS2", hp);
    nbout = wp * bytepix;
    nbimage = nbout * hp;
    nblock = nbimage / 2880;
    nbleft = (nblock + 1) * 2880 - nbimage;
    if (nbleft > 2880)
	nbleft = 0;

    if (icol1 > 1 | irow1 > 1) {
	hputs (header, "CTYPE1", "PIXEL");
	hputi4 (header, "CRPIX1", 1);
	hputi4 (header, "CRVAL1", ifcol1);
	hputi4 (header, "CDELT1", 1);
	hputs (header, "CTYPE2", "PIXEL");
	hputi4 (header, "CRPIX2", 1);
	hputi4 (header, "CRVAL2", ifrow1);
	hputi4 (header, "CDELT2", 1);
	}

    /* Write scaling keywords to get 16 bits of unsigned information */
    if (bytepix == 2 && !bitshift) {
	hputi4 (header, "BZERO", 32767);
	hputi4 (header, "BSCALE", 1);
	}

    /* Add HISTORY notice of this conversion */
    strcpy (history, RevMsg);
    endchar = strchr (history, ',');
    *endchar = (char) 0;
    strcat (history, " ");
    ltime = lt2fd ();
    strcat (history, ltime);
    endchar = strrchr (history,':');
    *endchar = (char) 0;
    strcat (history, " FITS from ");
    if (color == TIFF_RED)
	strcat (history, "red of ");
    else if (color == TIFF_GREEN)
	strcat (history, "green of ");
    else if (color == TIFF_BLUE)
	strcat (history, "blue of ");
    else
	strcat (history, "mean R+G+B of ");
    strcat (history, filename);
    if (bitshift)
	strcat (history, " shifted right 1 bit");
    if (strlen (history) > 72)
	history[72] = 0;
    hputc (header, "HISTORY", history);
    if (xcpix && ycpix) {
	strcpy (history, RevMsg);
	endchar = strchr (history, ',');
	*endchar = (char) 0;
	strcat (history, " ");
	ltime = lt2fd ();
	strcat (history, ltime);
	endchar = strrchr (history,':');
	*endchar = (char) 0;
	sprintf (temp, " %d x %d centered on [%d,%d]",
		 xdpix, ydpix, xcpix, ycpix);
	strcat (history, temp);
	hputc (header, "HISTORY", history);
	}
    else if (crange && rrange) {
	strcpy (history, RevMsg);
	endchar = strchr (history, ',');
	*endchar = (char) 0;
	strcat (history, " ");
	ltime = lt2fd ();
	strcat (history, ltime);
	endchar = strrchr (history,':');
	*endchar = (char) 0;
	sprintf (temp, " Rows %d - %d, Columns %d - %d",
		 ifrow1, ifrow2, ifcol1, ifcol2);
	strcat (history, temp);
	hputc (header, "HISTORY", history);
	}

    /* If assigned output name, use it */
    if (outname[0] != (char) 0)
	strcpy (fitsname, outname);

    /* Write FITS file to a specified directory */
    if (outdir[0] > 0) {
	strcpy (fitspath, outdir);
	strcat (fitspath, "/");
	fitsfile = strrchr (fitsname,'/');
	if (fitsfile == NULL)
	    strcat (fitspath, fitsname);
	else
	    strcat (fitspath, fitsfile+1);
	}
    else
	strcpy (fitspath, fitsname);
    first = 0;

    /* Write FITS image */
    nbrow = ncols * bytepix;
    if (sampix > 1)
	nbrow = nbrow * sampix;
    nplast = (icol1 + wp) * sampix;
    npfrst = (icol1 - 1) * sampix;
    if ((fdw = fitswhead (fitspath, header)) > 2) {
	if (verbose)
	    fprintf (stderr, "\n  to FITS file %s\n", fitspath);
	buf = _TIFFmalloc(TIFFScanlineSize(tif));

	/* Set up output */
	bufout = (short *) calloc (ncols, 2);

	/* Read TIFF data and translate to FITS */
	for (irow = irow2; irow >= irow1; irow--) {
	    row = irow;
	    TIFFReadScanline(tif, buf, row, 0);
	    if (sampix > 1) {
		if (bytepix == 2) {
		    bufin2 = (ushort *) buf + sampix * (icol1 - 1);
		    bufout2 = (ushort *) bufout;
		    switch (color) {
		    case TIFF_RED:
			for (i = nplast-3; i >= npfrst; i = i - sampix) {
			    *bufout2 = bufin2[i];
			    bufout2++;
			    }
			break;
		    case TIFF_GREEN:
			for (i = nplast-2; i >= npfrst; i = i - sampix) {
			    *bufout2 = bufin2[i];
			    bufout2++;
			    }
			break;
		    case TIFF_BLUE:
			for (i = nplast-1; i >= npfrst; i - sampix) {
			    *bufout2 = bufin2[i];
			    bufout2++;
			    }
			break;
		    case TIFF_RG:
			for (i = nplast-3; i >= npfrst; i = i - sampix) {
			    tpix = 0;
			    for (j = 0; j < 2; j++)
				tpix = tpix + (int) bufin2[i+j];
			    *bufout2 = (char) (tpix / 2);
			    bufout2++;
			    }
			break;
		    case TIFF_RB:
			for (i = nplast-3; i >= npfrst; i = i - sampix) {
			    tpix = 0;
			    for (j = 0; j < 3; j=j+2)
				tpix = tpix + (int) bufin2[i+j];
			    *bufout2 = (char) (tpix / 2);
			    bufout2++;
			    }
			break;
		    case TIFF_GB:
			for (i = nplast-3; i >= npfrst; i = i - sampix) {
			    tpix = 0;
			    for (j = 1; j < 3; j++)
				tpix = tpix + (int) bufin2[i+j];
			    *bufout2 = (char) (tpix / 2);
			    bufout2++;
			    }
			break;
		    case TIFF_RGB:
			for (i = nplast-3; i >= npfrst; i = i - sampix) {
			    tpix = 0;
			    for (j = 0; j < sampix; j++)
				tpix = tpix + (int) bufin2[i+j];
			    *bufout2 = (short) (tpix / 3);
			    bufout2++;
			    }
			break;
			}

		    /* If image is a negative, convert it to a positive */
		    if (negative) {
			bufout2 = (ushort *) bufout;
			if (limitval) {
			    for (i = 0; i < wp; i++) {
				if (bufout2[i] > limitval)
				    itemp = limitval;
				else {
				    itemp = 65535 - bufout2[i];
				    itemp = itemp + background;
				    }
				if (itemp < 0)
				    bufout2[i] = 0;
				else
				    bufout2[i] = (ushort) itemp;
				}
			    }
			else if (background) {
			    for (i = 0; i < wp; i++) {
				itemp = (65535 - bufout2[i]) + background;
				if (itemp < 0)
				    bufout2[i] = 0;
				else
				    bufout2[i] = (ushort) itemp;
				}
			    }
			else {
			    for (i = 0; i < wp; i++) {
				itemp = 65535 - bufout2[i];
				if (itemp < 0)
				    bufout2[i] = 0;
				else
				    bufout2[i] = (ushort) itemp;
				}
			    }
			}
		    }

		/* Deal with 1-byte TIFF images */
		else {
		    bufin1 = ((char *) buf) + sampix * (icol1 - 1);
		    bufout1 = (char *) bufout;
		    switch (color) {
		    case TIFF_RED:
			for (i = nplast; i >= npfrst; i = i - sampix) {
			    *bufout1 = bufin1[i];
			    bufout1++;
			    }
			break;
		    case TIFF_GREEN:
			for (i = nbrow-2; i >= npfrst; i = i - sampix) {
			    *bufout1 = bufin1[i];
			    bufout1++;
			    }
			break;
		    case TIFF_BLUE:
			for (i = nbrow-1; i >= npfrst; i = i - sampix) {
			    *bufout1 = bufin1[i];
			    bufout1++;
			    }
			break;
		    case TIFF_RG:
			for (i = nplast; i >= npfrst; i = i - sampix) {
			    tpix = npfrst;
			    for (j = npfrst; j < 2; j++)
				tpix = tpix + (int) bufin1[i+j];
			    *bufout1 = (char) (tpix / 2);
			    bufout1++;
			    }
			break;
		    case TIFF_RB:
			for (i = nplast; i >= npfrst; i = i - sampix) {
			    tpix = npfrst;
			    for (j = npfrst; j < 3; j=j+2)
				tpix = tpix + (int) bufin1[i+j];
			    *bufout1 = (char) (tpix / 2);
			    bufout1++;
			    }
			break;
		    case TIFF_GB:
			for (i = nplast; i >= npfrst; i = i - sampix) {
			    tpix = npfrst;
			    for (j = 1; j < 3; j++)
				tpix = tpix + (int) bufin1[i+j];
			    *bufout1 = (char) (tpix / 2);
			    bufout1++;
			    }
			break;
		    case TIFF_RGB:
			for (i = nplast; i >= npfrst; i = i - sampix) {
			    tpix = npfrst;
			    for (j = npfrst; j < sampix; j++)
				tpix = tpix + (int) bufin1[i+j];
			    *bufout1 = (char) (tpix / 3);
			    bufout1++;
			    }
			break;
			}

		    /* If negative, convert image to positive */
		    if (negative) {
			bufout1 = (char *) bufout;
			if (background) {
			    for (i = npfrst; i < wp; i++) {
				bufout1[i] = (256 - bufout1[i]) + background;
				if (bufout1[i] < 0)
				    bufout1[i] = 0;
				}
			    }
			else {
			    for (i = 0; i < wp; i++) {
				itemp =  256 - (int) bufout1[i];
				if (itemp < 0)
				    bufout1[i] = 0;
				else
				    bufout1[i] = (char) itemp;
				}
			    }
			}
		    }
		}

	    /* Convert negative to positive if single sample per pixel */
	    else if (negative) {
		if (bytepix == 2) {
		    bufin2 = (ushort *) buf + icol1 - 1;
		    bufout2 = (ushort *) bufout;
		    if (limitval) {
			for (i = wp-1; i >= 0; i--) {
			    if (bufin2[i] > limitval)
				itemp = limitval;
			    else {
				itemp = 65535 - bufin2[i];
				itemp = itemp + background;
				}
			    if (itemp < 0)
				*bufout2 = 0;
			    else
				*bufout2 = (ushort) itemp;
			    bufout2++;
			    }
			}
		    else if (background) {
			for (i = wp-1; i >= 0; i--) {
			    itemp = (65535 - bufin2[i]) + background;
			    if (itemp < 0)
				*bufout2 = 0;
			    else
				*bufout2 = (ushort) itemp;
			    bufout2++;
			    }
			}
		    else {
			for (i = wp-1; i >= 0; i--) {
			    itemp = 65535 - bufin2[i];
			    if (itemp < 0)
				*bufout2 = 0;
			    else
				*bufout2 = (ushort) itemp;
			    bufout2++;
			    }
			}
		    }
		else {
		    bufin1 = (char *) buf + icol1 - 1;
		    bufout1 = (char *) bufout;
		    for (i = wp-1; i >= 0; i--) {
			itemp = 256 - (int) bufin1[i];
			if (itemp < 0)
			    *bufout1 = 0;
			else
			    *bufout1 = (char) itemp;
			bufout1++;
			}
		    }
		}

	    /* At a minimum, reverse each row */
	    else {
		if (bytepix == 2) {
		    bufin2 = (ushort *) buf + icol1 - 1;
		    bufout2 = (ushort *) bufout;
		    for (i = wp-1; i >= 0; i--) {
			*bufout2 = bufin2[i];
			bufout2++;
			}
		    }
		else {
		    bufin1 = (char *) buf + icol1 - 1;
		    bufout1 = (char *) bufout;
		    for (i = wp-1; i >= 0; i--) {
			*bufout1 = bufin1[i];
			bufout1++;
			}
		    }
		}

	    /* Write out this row, and log it, if it is time */
	    if (bytepix == 2) {
		bufout2s = bufout;
		bufout2u = (ushort *) bufout;
		if (bitshift) {
		    for (i = 0; i < wp; i++) {
			*bufout2s++ = *bufout2u++ >> 1;
			}
		    }
		else {
		    for (i = 0; i < wp; i++) {
			vpix = (int) *bufout2u++;
			vpix = vpix - 32767;
			*bufout2s++ = (short) vpix;
			}
		    }
		write (fdw, bufout, nbout);
		}
	    else
		write (fdw, bufout1, nbout);
	    if (nlog > 0 && row % nlog == 0) {
		fprintf (stderr, "Row %5d operations complete   ", row);
		(void) putc (13,stderr);
		}
            }

	/* Pad last FITS block to 2880 */
	if (nbleft > 0) {
	    for (i = 0; i < nbleft; i++)
		write (fdw, &ib0, 1);
	    }

	/* Close input file and free buffers */
	_TIFFfree(buf);
	_TIFFfree(bufout);
	TIFFClose(tif);
	if (verbose)
	    fprintf (stderr, "\n%d rows of image written         \n", hp);
	}

    else if (verbose)
	fprintf (stderr, "\n%s: not written.\n", fitspath);

    fprintf (stderr, "\n");
    free (header);
    free (image);
    return;
}


/* RANGEINIT -- Initialize range structure from string */

static struct Range *
RangeInit (string, ndef)

char	*string;	/* String containing numbers separated by , and - */
int	ndef;		/* Maximum allowable range value */

{
    struct Range *range;
    int ip, irange;
    char *slast;
    double first, last, step;

    if (!isrange (string) && !isnum (string))
	return (NULL);
    ip = 0;
    range = (struct Range *)calloc (1, sizeof (struct Range));
    range->irange = -1;
    range->nvalues = 0;
    range->nranges = 0;

    for (irange = 0; irange < MAXRANGE; irange++) {

	/* Default to entire list */
	first = 1.0;
	last = ndef;
	step = 1.0;

	/* Skip delimiters to start of range */
	while (string[ip] == ' ' || string[ip] == '	' ||
	       string[ip] == ',')
	    ip++;

	/* Get first limit
	 * Must be a number, '-', 'x', or EOS.  If not return ERR */
	if (string[ip] == (char)0) {	/* end of list */
	    if (irange == 0) {

		/* Null string defaults */
		range->ranges[0] = first;
		if (first < 1)
		    range->ranges[1] = first;
		else
		    range->ranges[1] = last;
		range->ranges[2] = step;
		range->nvalues = range->nvalues + 1 +
			  ((range->ranges[1]-range->ranges[0])/step);
		range->nranges++;
		return (range);
		}
	    else
		return (range);
	    }
	else if (string[ip] > (char)47 && string[ip] < 58) {
	    first = strtod (string+ip, &slast);
	    ip = slast - string;
	    }
	else if (strchr ("-:x", string[ip]) == NULL) {
	    free (range);
	    return (NULL);
	    }

	/* Skip delimiters */
	while (string[ip] == ' ' || string[ip] == '	' ||
	       string[ip] == ',')
	    ip++;

	/* Get last limit
	* Must be '-', or 'x' otherwise last = first */
	if (string[ip] == '-' || string[ip] == ':') {
	    ip++;
	    while (string[ip] == ' ' || string[ip] == '	' ||
	   	   string[ip] == ',')
		ip++;
	    if (string[ip] == (char)0)
		last = first + ndef;
	    else if (string[ip] > (char)47 && string[ip] < 58) {
		last = strtod (string+ip, &slast);
		ip = slast - string;
		}
	    else if (string[ip] != 'x')
		last = first + ndef;
	    }
	else if (string[ip] != 'x')
	    last = first;

	/* Skip delimiters */
	while (string[ip] == ' ' || string[ip] == '	' ||
	       string[ip] == ',')
	    ip++;

	/* Get step
	 * Must be 'x' or assume default step. */
	if (string[ip] == 'x') {
	    ip++;
	    while (string[ip] == ' ' || string[ip] == '	' ||
	   	   string[ip] == ',')
		ip++;
	    if (string[ip] == (char)0)
		step = 1.0;
	    else if (string[ip] > (char)47 && string[ip] < 58) {
		step = strtod (string+ip, &slast);
		ip = slast - string;
		}
	    else if (string[ip] != '-' && string[ip] != ':')
		step = 1.0;
            }

	/* Output the range triple */
	range->ranges[irange*3] = first;
	range->ranges[irange*3 + 1] = last;
	range->ranges[irange*3 + 2] = step;
	range->nvalues = range->nvalues + ((last-first+(0.1*step)) / step + 1);
	range->nranges++;
	}

    return (range);
}


/* ISRANGE -- Return 1 if string is a range, else 0 */

static int
isrange (string)

char *string;		/* String which might be a range of numbers */

{
    int i, lstr;

    /* If range separators present, check to make sure string is range */
    if (strchr (string+1, '-') || strchr (string+1, ',')) {
	lstr = strlen (string);
	for (i = 0; i < lstr; i++) {
	    if (strchr ("0123456789-,.x", (int)string[i]) == NULL)
		return (0);
	    }
	return (1);
	}
    else
	return (0);
}


/* RSTART -- Restart at beginning of range */

static void
rstart (range)

struct Range *range;	/* Range structure */

{
    range->irange = -1;
    return;
}


/* RGETN -- Return number of values from range structure */

static int
rgetn (range)

struct Range *range;	/* Range structure */

{
    return (range->nvalues);
}


/*  RGETR8 -- Return next number from range structure as 8-byte f.p. number */

static double
rgetr8 (range)

struct Range *range;	/* Range structure */

{
    int i;

    if (range == NULL)
	return (0.0);
    else if (range->irange < 0) {
	range->irange = 0;
	range->first = range->ranges[0];
	range->last = range->ranges[1];
	range->step = range->ranges[2];
	range->value = range->first;
	}
    else {
	range->value = range->value + range->step;
	if (range->value > range->last) {
	    range->irange++;
	    if (range->irange < range->nranges) {
		i = range->irange * 3;
		range->first = range->ranges[i];
		range->last = range->ranges[i+1];
		range->step = range->ranges[i+2];
		range->value = range->first;
		}
	    else
		range->value = 0.0;
	    }
	}
    return (range->value);
}


/*  RGETI4 -- Return next number from range structure as 4-byte integer */

static int
rgeti4 (range)

struct Range *range;	/* Range structure */

{
    double value;

    value = rgetr8 (range);
    return ((int) (value + 0.000000001));
}

/* Jan 14 2002	New program
 * Jan 15 2002	Always write in current working directory
 * Jan 15 2002	Deal with multiple samples per pixel and 2-byte samples
 * Jan 16 2002	Add options to convert negatives and use one of three colors
 * Mar 28 2002	Add options to use 2/3 of colors
 * Apr 18 2002	Change -k to add constant (which may be negative)
 * Apr 18 2002	Set pixel values greater than -l value to it
 * Jul 19 2002	For 16-bit out, convert unsigned to signed using BZERO,BSCALE
 * Aug 30 2002	Add -1 option to shift right instead of scaling
 * Oct  4 2002	Add -c and -x options to extract an image section
 * Oct 10 2002	Add x and y range option to extract an image section
 * Oct 10 2002	Always set bitshift for 2-byte conversions (TIFF is unsigned)
 */
